#include  <limits.h>
#include <image.h>
#include <assert.h>
#include "colormap.h"
#include "inverse.h"

#include <glib.h>
#include "tab_int.h"

#define ABS(a) ((a)<0?-(a):(a))

static float alpha=0.5;
static float minus_alpha=0.5;

static GHashTable *hash_rep;
static tab_int set_int;
static int hash_nb_color;

static int comp;
static double* nearest_dist;

typedef bouleen (*PFcolormap_next)(colormap);
static PFcolormap_next colormap_next=image_pixel_droite;
static PFcolormap_next colormap_previous=image_pixel_gauche;

typedef int (*PFbasic_mapping)(colormap,int*);

static int map_hash_priv(colormap, int*);
static int map_trivial_priv(colormap, int*);
static int map_partial_priv(colormap, int*);
static int map_L_alpha_priv(colormap, int*);
static int map_sorted_priv(colormap,int*);
static int map_nearest_priv(colormap, int*);

static void get_nearest_priv(colormap);
static int   dicho_search_get_index_priv(colormap,int);
static image map_image_priv(image,colormap,PFbasic_mapping);
static void  sorted_search_in_one_direction_priv(colormap,
						 int*, /* input color */
						 int, /* start index  */
						 int* , /* min index */
						 long*, /* min */
						 PFcolormap_next); /*
								     sense of search */

/*********** Public Functions **************/
image inverse_trivial(image input,colormap map)
{
  return map_image_priv(input,map,map_trivial_priv);
}
image inverse_partial(image input,colormap map)
{
  return map_image_priv(input,map,map_partial_priv);
}
image inverse_norm_alpha(image input,colormap map)
{
  return map_image_priv(input,map,map_L_alpha_priv);
}
image inverse_sorted(image input,colormap map)
{
  colormap sorted_map = colormap_give_sorted(map,3,&comp);  
  image output;

  output= map_image_priv(input,sorted_map,map_sorted_priv);
  DEFAIRE_colormap(sorted_map);

  return output;
}
image inverse_nearest(image input,colormap map)
{
  image output;

  get_nearest_priv(map);
  output= map_image_priv(input,map,map_nearest_priv);
  free(nearest_dist);

  return output;
}
image inverse_hash(image input,colormap map)
{
  image img;
  hash_rep=g_hash_table_new(g_int_hash,g_int_equal);
  set_int = FAIRE_tab_int();
  hash_nb_color=0;

  img= map_image_priv(input,map,map_hash_priv);

  g_hash_table_destroy(hash_rep);
  DEFAIRE_tab_int(set_int);

  return img;
}

void inverse_set_alpha(float a)
{
  assert(a>=0.0 && a<=1.0);

  alpha=a;
  minus_alpha=1.0-a;
}
/************** Private Functions ******************/
static int map_trivial_priv(colormap map, int* input)
{
  int* pixel;
  long dist;
  long min_dist=LONG_MAX;
  int index;
  int i=0;

  colormap_begin(map);
  
 do
   {
     pixel=colormap_read_color(map);
     dist = (input[0]-pixel[0])*(input[0]-pixel[0])+
       (input[1]-pixel[1])*(input[1]-pixel[1])+
       (input[2]-pixel[2])*(input[2]-pixel[2]);
     if(dist < min_dist)
       {
	 index=i;
	 min_dist=dist;
       }
     i++;
   }
 while(colormap_next(map));

 return index;
}
static int map_L_alpha_priv(colormap map, int* input)
{
  int* pixel;
  long dist,dist_inf,dist_1,tmp;
  long min_dist=LONG_MAX;
  int index;
  int i=0;
  
  colormap_begin(map);
  
 do
   {
     pixel=colormap_read_color(map);
     dist_1 = dist_inf=ABS(input[0]-pixel[0]);

     tmp=ABS(input[1]-pixel[1]);
     if(tmp>dist_inf)
       dist_inf=tmp;
     dist_1+=tmp;

     tmp=ABS(input[2]-pixel[2]);
     if(tmp>dist_inf)
       dist_inf=tmp;
     dist_1+=tmp;

     if(alpha==0.5)
       dist = (dist_1+dist_inf)>>1;
     else
       dist = minus_alpha*dist_1+alpha*dist_inf;
     if(dist < min_dist)
       {
	 index=i;
	 min_dist=dist;
       }
     i++;
   }
 while(colormap_next(map));

 return index;
}
static int map_partial_priv(colormap map, int* input)
{
  int* pixel;
  long dist;
  long min_dist=LONG_MAX;
  long min_dist_global=LONG_MAX;
  int index;
  int i=0;

  colormap_begin(map);
  
 do
   {
     pixel=colormap_read_color(map);
     dist = (input[0]-pixel[0])*(input[0]-pixel[0]);
     if(dist>min_dist)
       {
	 i++;
	 continue;
       }
     dist+=(input[1]-pixel[1])*(input[1]-pixel[1]);
     if(dist>min_dist)
       {
	 i++;
	 continue;
       }
     dist+=  (input[2]-pixel[2])*(input[2]-pixel[2]);

     if(dist < min_dist)
       {
	 index=i;
	 min_dist=dist;
       }
     i++;
   }
 while(colormap_next(map));

 return index;
}

image map_image_priv(image input,colormap map,PFbasic_mapping mapping)
{
  int* pixel;
  int color[3];
  int i,index;
  

  image output= FAIRE_image();
  image_initialize(output,
		   3,
		   image_give_largeur(input),
		   image_give_hauteur(input)
		   );
  image_debut(input);
  image_debut(output);
  do
    {
      pixel=image_lire_pixel(input);
      for(i=0;i<3;i++)
	color[i]=pixel[i];

      index=mapping(map,color);

      colormap_move_to(map,index);
      image_ecrire_pixel(output,
			 colormap_read_color(map));

      image_pixel_suivant(output);
      
    }
  while(image_pixel_suivant(input));

  return output;
}
static int dicho_search_get_index_priv(colormap map,int value)
{
  int min=0;
  int mid,max=colormap_give_nb_color(map)-1;
  int map_value,max_value,min_value;

  while (max-min != 1)
    {
      mid=(max+min)/2;
      map_value=colormap_give_color_comp(map,mid,comp);
      if(map_value == value)
	return mid;
      if(map_value < value)
	min=mid;
      else
	max=mid;
    }

  min_value=colormap_give_color_comp(map,min,comp);
  max_value=colormap_give_color_comp(map,max,comp);
  if(value-min_value < max_value-value)
    return min;
  return max;
}
static void sorted_search_in_one_direction_priv(colormap map,
						int* input,
						int index_start,
						int* min_ind,
						long* min,
						PFcolormap_next next)
{
  int* pixel;
  long dist;
  long min_dist=LONG_MAX;
  int index;
  int j,i=0;
  int axis[3];
  
  axis[0]=comp;
  axis[1]=(comp+1)%3;
  axis[2]=(comp+2)%3;
  
  colormap_move_to(map,index_start);
  do
    {
      pixel=colormap_read_color(map);
      dist = (input[comp]-pixel[comp])*(input[comp]-pixel[comp]);
      if(dist>min_dist)
	{
	  *min_ind=index;
	  *min=min_dist;
	  return;
	}
      for(j=1;j<3;j++)
	dist+=(input[axis[j]]-pixel[axis[j]])*(input[axis[j]]-pixel[axis[j]]);
      
      if(dist < min_dist)
	{
	  index=i;
	  min_dist=dist;
	}
      i++;
    }
  while(next(map));
  
  *min_ind=index;
  *min=min_dist;
  
}
static int map_sorted_priv(colormap map,int* input)
{
  int index_start;
  int offset1,offset2;
  long min1,min2;


  index_start=dicho_search_get_index_priv(map,input[comp]);
  sorted_search_in_one_direction_priv(map,
				      input,
				      index_start,
				      &offset1,&min1,
				      colormap_next);
  
  sorted_search_in_one_direction_priv(map,
				      input,
				      index_start,
				      &offset2,&min2,
				      colormap_previous);
  

  
  if(min1<min2)
    return index_start+offset1;
  
  return index_start-offset2;

}
static int map_nearest_priv(colormap map, int* input)
{
  int* pixel;
  long dist;
  long min_dist=LONG_MAX;
  int index;
  int i=0;


  colormap_begin(map);
  
 do
   {
     pixel=colormap_read_color(map);
     dist = (input[0]-pixel[0])*(input[0]-pixel[0])+
       (input[1]-pixel[1])*(input[1]-pixel[1])+
       (input[2]-pixel[2])*(input[2]-pixel[2]);     
     if(dist < min_dist)
       {
	 index=i;
	 min_dist=dist;
	 
	 if(sqrt(dist) <= 0.5*nearest_dist[i])
	   return i;
       }
     i++;
   }
 while(colormap_next(map));

 return index;
}
static int map_hash_priv(colormap map, int* input)
{
  int cle= input[0]<<16|input[1]<<8|input[2];
  int* ptr_cle,*ptr_rep;

  if(!g_hash_table_lookup_extended(hash_rep,
				    (gconstpointer)&cle,
				    (gpointer)&ptr_cle,
				   (gpointer)&ptr_rep) )
  {
    ptr_cle=tab_int_give_elt(set_int,hash_nb_color);
    *ptr_cle=cle;
    hash_nb_color++;
    ptr_rep=tab_int_give_elt(set_int,hash_nb_color);
    *ptr_rep =map_trivial_priv(map,input);
    hash_nb_color++;
    g_hash_table_insert(hash_rep,(gpointer)ptr_cle,(gpointer)ptr_rep);

    return *ptr_rep;
  }
     return *ptr_rep;
}
static void get_nearest_priv(colormap map)
{
  int nb_color = colormap_give_nb_color(map);
  int* color;
  int i,j,k;
  int comp_tmp;
  long dist,min_dist;
  double * tmp_dist;

  nearest_dist = (double*)malloc(nb_color*sizeof(double));
  tmp_dist=nearest_dist;
  
  colormap_begin(map);
  i=0;
  do
    {      
      color=colormap_read_color(map);
      min_dist=LONG_MAX;
      for(j=0;j<nb_color;j++)
	{
	  if(i==j)
	    continue;
	  dist=0;
	  for(k=0;k<3;k++)
	    {
	      comp_tmp=colormap_give_color_comp(map,j,k);
	      dist+=(color[k]-comp_tmp)*(color[k]-comp_tmp);
	    }
	  if(dist<min_dist)
	    min_dist=dist;
	}
      *tmp_dist=sqrt(min_dist);
      tmp_dist++;
      i++;
    }
  while(colormap_next(map));
}
      
