#include <libgen.h>
#include <math.h>
#include <image.h>
#include "inverse.h"
#include "colormap.h"
#include <sys/types.h>
#include <sys/times.h>
#include <limits.h>
#include <time.h>

#define TRIVIAL 0
#define PARTIAL 1
#define ALPHA   2
#define SORT    3
#define NEAREST 4
#define HASH    5


void usage(char * cmd, int  code)
{
  fprintf(stderr, "Usage: %s |-trivial][-L_alpha [-alpha alpha]][-partial][-sort][-nearest][-hash] input.ppm map.pp\n",cmd);
  
  exit(code);
}
int get_method(char* cmd,char * method)
{
  static char* method_names[]={"-trivial",
			       "-partial",
			       "-l_alpha",
			       "-sort",
			       "-nearest",
			       "-hash"};
  static int nb_methods =6;
  int i;

  for(i=0;i<strlen(method);i++)
    method[i]=tolower(method[i]);

  for(i=0;i<nb_methods;i++)
    {
      if (strcmp(method,method_names[i]) == 0)
	return i;
    }
  usage(cmd,4);
}
int get_alpha(char* cmd,char * argv[],int* argc)
{
  int i,j;
  char arg[80];
  int nb_arg=*argc;
  char *convert=NULL;
  float alpha;
  
  arg[0]='-';
  for(i=1;i<nb_arg;i++)
    {
      if(argv[i][0] != '-')
	continue;
      
      for(j=1;j<strlen(argv[i]);j++)
	arg[j]=tolower(argv[i][j]);
      arg[strlen(argv[i])]='\0';

      if (strcmp(arg,"-alpha") == 0)
	break;
    }
  if(i== nb_arg)
    return 0;

  alpha=strtod(argv[i+1],&convert);
  if(alpha == HUGE_VAL || convert== argv[i+1])
    usage(cmd,5);

  inverse_set_alpha(atof(argv[i+1]));
  if(i== nb_arg-3)
    argv[i]=argv[nb_arg-1];
  
  for(j=i;j<nb_arg-3;j+=2)
    {
      argv[j]=argv[j+2];
      argv[j+1]=argv[j+3];
    }
  *argc=nb_arg-2;

  return 1;
}

int main(int argc, char* argv[])
{
  image output,input;
  colormap map;
  int method;
  char* cmd=basename(argv[0]);
  struct tms tms_debut, tms_fin;
  long temps;
  get_alpha(cmd,argv,&argc);

  if(argc != 4)
    usage(cmd,1);
  
  method=get_method(argv[0],argv[1]);
  input= FAIRE_image();
  if (image_charger(input,argv[2]) != 0)
    {
      fprintf(stderr,"Unable to open input image %s exiting\n",argv[2]);
      DEFAIRE_image(input);
      exit(2);
    }
  map= FAIRE_colormap();
  if (colormap_load(map,argv[3]) != 0)
    {
      fprintf(stderr,"Unable to open colormap %s exiting\n",argv[3]);
      DEFAIRE_image(input);
      DEFAIRE_image(map);
      exit(3);
    }
  times(&tms_debut);
  output = inverse[method](input,map);
  times(&tms_fin);
  temps=tms_fin.tms_utime - tms_debut.tms_utime;
  image_to_stream(output,stdout);
  fprintf(stderr,"Temps: %f\n",((float)temps)/CLK_TCK);
  DEFAIRE_image(input);
  DEFAIRE_image(output);
  DEFAIRE_image(map);

}
